import numpy as np
from numpy import ndarray
from rasterstats import zonal_stats
from geopandas import GeoSeries
from affine import Affine
from typing import List, Union
import xarray
import pandas as pd
from pathlib import Path
from datetime import date


def flatten_stats(stats_dict: List[str], stats: List[str]) -> List[float]:
    return np.array([[item[stat] for stat in stats] for item in stats_dict]).flatten()


def sample_geoseries(
    values: ndarray,
    geometries: Union[List, GeoSeries],
    affine: Affine,
    nodata: float,
    stats: Union[str, List[str]] = "mean",
) -> List[float]:

    # we work with lists to flatten_stats
    if isinstance(stats, str):
        stats = [stats]

    stats_dict = zonal_stats(
        geometries,
        values,
        affine=affine,
        nodata=nodata,
        stats=" ".join(stats),
        boundless=True,
    )

    return flatten_stats(stats_dict, stats)


def sample_netcdf(
    nc_file: Path | str,
    variable_code: str,
    geometries: Union[List, GeoSeries],
    stats: Union[str, List[str]] = "mean",
    start_date: date | None = None,
    end_date: date | None = None,
    unlink: bool = False,
) -> pd.DataFrame:
    """Sample a set of geometries over a netcdf file

    Parameters
    ----------
    nc_file : Path or str
        path to NetCDF file
    variable_code : str
        Variable in NetCDF file to sample
    geometries : Union[List, GeoSeries]
        geometries to sample
    stats : List[str]
        statistics to sample
    start_date : date | None
        start date for selection, by default None
    end_date: date | None
        end date for selection, by default None
    unlink : bool, optional
        option to delete netcdf-file after sampling, by default False

    Returns
    -------
    pd.DataFrame
        Pandas DataFrame with statistics per timestamp per geometry
    """
    if isinstance(nc_file, str):
        nc_file = Path(str)
    assert nc_file.is_file(), f"nc_file {nc_file} does not exist"

    # read temp-source for sampling
    with xarray.open_dataset(nc_file, engine="netcdf4") as ds:
        if (start_date is not None) and (end_date is not None):
            ds = ds.sel(time=slice(start_date, end_date))
        nodata = ds[variable_code].attrs.get("_FillValue")
        affine = ds.rio.transform()
        data = {
            time: sample_geoseries(
                values=ds[variable_code].sel(time=time).values,
                geometries=geometries,
                affine=affine,
                nodata=nodata,
                stats=stats,
            )
            for time in ds["time"].values
        }

    # delete temp-file
    if unlink:
        if nc_file.exists():
            nc_file.unlink()

    # create columns
    if isinstance(stats, str):
        stats = [stats]

    if len(stats) == 1:
        columns = geometries.index
    else:
        geom_index = geometries.index
        columns = pd.MultiIndex.from_product(
            iterables=[geom_index, stats], names=["index", "stats"]
        )

    return pd.DataFrame.from_dict(data, orient="index", columns=columns)


def sample_netcdfs(
    nc_files: list[Path],
    variable_code: str,
    geometries: Union[List, GeoSeries],
    stats: Union[str, List[str]] = "mean",
    start_date: date | None = None,
    end_date: date | None = None,
):
    """Sample over a set of netcdf-files

    Parameters
    ----------
    nc_files : list[Path]
        A list of netcdf-files
    variable_code : str
        Variable in NetCDF file to sample
    geometries : Union[List, GeoSeries]
        geometries to sample
    stats : List[str]
        statistics to sample
    start_date : date | None
        start date for selection, by default None
    end_date: date | None
        end date for selection, by default None

    Returns
    ------
    pd.DataFrame
        Pandas DataFrame with statistics per timestamp per geometry
    """
    assert nc_files, f"no NetCDF files to sample"
    for nc_file in nc_files:
        assert nc_file.is_file(), f"nc_file {nc_file} does not exist"

    # read all transforms to see if dataset is consistent
    transforms = list(
        set(xarray.open_dataset(file).rio.transform() for file in nc_files)
    )
    if len(transforms) == 1:
        with xarray.open_dataset(nc_files[0], decode_coords="all") as ds:
            geometries = geometries.to_crs(ds.rio.crs)
    else:
        raise ValueError(
            f"Files do not have one consistent transform. Got {transforms}"
        )

    dfs = [
        sample_netcdf(
            nc_file,
            variable_code,
            geometries,
            stats=stats,
            start_date=start_date,
            end_date=end_date,
            unlink=False,
        )
        for nc_file in nc_files
    ]
    dfs = [i for i in dfs if not i.empty]
    df = pd.concat(dfs).sort_index()

    return df


def sample_nc_dir(
    dir_path: Path | str,
    variable_code: str,
    geometries: Union[List, GeoSeries],
    stats: Union[str, List[str]] = "mean",
    start_date: date | None = None,
    end_date: date | None = None,
):
    """Sample over a directory of netcdf-files

    Parameters
    ----------
    dir_path : Path | str
        Directory with netcdf files
    variable_code : str
        Variable in NetCDF file to sample
    geometries : Union[List, GeoSeries]
        geometries to sample
    stats : List[str]
        statistics to sample
    start_date : date | None
        start date for selection, by default None
    end_date: date | None
        end date for selection, by default None

    Returns
    ------
    pd.DataFrame
        Pandas DataFrame with statistics per timestamp per geometry
    """
    if isinstance(dir_path, str):
        dir_path = Path(dir_path)
    assert dir_path.is_dir(), f"dir_path {dir_path} does not exist"

    nc_files = [x for x in dir_path.iterdir() if x.suffix == ".nc"]

    df = sample_netcdfs(
        nc_files,
        variable_code=variable_code,
        geometries=geometries,
        stats=stats,
        start_date=start_date,
        end_date=end_date,
    )
    return df



import logging
from typing import List
import requests
from wiwb.api_calls import Request
from dataclasses import dataclass, field


logger = logging.getLogger(__name__)


@dataclass
class GetVariables(Request):
    """GetVariables request"""

    data_source_codes: List[str] = field(default_factory=list)
    variable_codes: List[str] = field(default_factory=list)

    @property
    def url_post_fix(self) -> str:
        return "entity/variables/get"

    @property
    def json(self) -> dict:
        return {
            "DataSourceCodes": self.data_source_codes,
            "VariableCodes": self.variable_codes,
        }

    def run(self) -> List[str]:
        response = requests.post(self.url, headers=self.auth.headers, json=self.json)

        if response.ok:  # return list of data sources
            return response.json()["Variables"]

        else:  # raise Error
            response.raise_for_status()



__version__ = "2024.4.1"
from wiwb.api import Api
from wiwb.auth import Auth
import warnings

__all__ = ["Auth", "Api"]

warnings.filterwarnings(
    "ignore",
    category=DeprecationWarning,
    module="dateutil",
    message="Use timezone-aware objects to represent datetimes in UTC",
)
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    module="pyproj",
    message="You will likely lose important projection information when converting to a PROJ string from another",
)
warnings.filterwarnings(
    "ignore",
    category=UserWarning,
    module="rasterstats",
    message="Setting nodata to -999; specify nodata explicitly",
)



import requests
import logging
from typing import Dict
from wiwb.api_calls import Request
from dataclasses import dataclass, field
from wiwb.constants import PRIMARY_STRUCTURE_TYPES

logger = logging.getLogger(__name__)


@dataclass
class GetDataSources(Request):
    """GetDataSources request"""

    primary_structure_types: PRIMARY_STRUCTURE_TYPES = field(
        default_factory=lambda: ["Grid"]
    )

    @property
    def url_post_fix(self) -> str:
        return "entity/datasources/get"

    def run(self) -> Dict:
        response = requests.post(self.url, headers=self.auth.headers, json={})

        if response.ok:  # return list of data sources
            return {
                k: v
                for k, v in response.json()["DataSources"].items()
                if v["PrimaryStructureType"] in self.primary_structure_types
            }

        else:  # raise Error
            response.raise_for_status()



import logging
from wiwb.api_calls import Request
from dataclasses import dataclass, field, InitVar
from typing import Iterable, List, Tuple, Union, Optional
from geopandas import GeoSeries
from pandas import DataFrame
from shapely.geometry import Point, Polygon, MultiPolygon
from datetime import date
from wiwb.converters import snake_to_pascal_case
import pyproj
import requests
from wiwb.sample import sample_netcdf
from pathlib import Path
import tempfile
from wiwb.constants import (
    get_defaults,
    FILE_SUFFICES,
    DATA_FORMAT_CODES,
    INTERVAL_TYPES,
)

logger = logging.getLogger(__name__)
defaults = get_defaults()


@dataclass
class Extent:
    f"""Extent for Settings in request body in correct epsg: {defaults.crs}.

    Parameters
    ----------
    xll : float
        The x-coordinate of the lower-left corner of the extent. Defaults to {defaults.bounds[0]}.
    yll : float
        The y-coordinate of the lower-left corner of the extent. Defaults to {defaults.bounds[1]}.
    xur : float
        The x-coordinate of the upper-right corner of the extent. Defaults to {defaults.bounds[2]}.
    yur : float
        The y-coordinate of the upper-right corner of the extent. Defaults to {defaults.bounds[3]}.
    """

    xll: float = defaults.bounds[0]
    yll: float = defaults.bounds[1]
    xur: float = defaults.bounds[2]
    yur: float = defaults.bounds[3]

    def __post_init__(self):
        if self.width <= 0:
            raise ValueError(
                f"'xll' ({self.xll}) should be smaller than 'xur' ({self.xur})"
            )

        if self.height <= 0:
            raise ValueError(
                f"'yll' ({self.yll}) should be smaller than 'yur' ({self.yur})"
            )

        self.correct_bounds()

    @property
    def width(self):
        return self.xur - self.xll

    @property
    def height(self):
        return self.yur - self.yll

    @property
    def crs(self):
        return pyproj.CRS(self.epsg)

    @property
    def epsg(self):
        return defaults.crs

    @property
    def spatial_reference(self):
        return {"Epsg": self.epsg}

    def correct_bounds(self):
        # get crs-unit
        units = None
        crs_dict = self.crs.to_dict()
        if "unit" in crs_dict.keys():
            units = crs_dict["units"]

        # get min width and height
        if units == "m":
            min_width_height = 10
        else:
            min_width_height = 0.0001  # we assume degrees

        # alter bounds
        if self.width < min_width_height:
            logger.warning(
                f"""Width of bounds < min_width ({self.width < min_width_height}). {self.xll} and {self.xur} will be adjusted"""  # noqa:E501
            )
            self.xll -= (min_width_height - self.width) / 2
            self.xur += (min_width_height - self.width) / 2

        if self.height < min_width_height:
            logger.warning(
                f"""Height of bounds < min_height ({self.height < min_width_height}). {self.yll} and {self.yur} will be adjusted"""  # noqa:E501
            )
            self.yll -= (min_width_height - self.height) / 2
            self.yur += (min_width_height - self.height) / 2

    def json(self):
        dict = self.__dict__.copy()
        dict["spatial_reference"] = self.spatial_reference
        return {snake_to_pascal_case(k): v for k, v in dict.items()}


@dataclass
class Interval:
    """Interval for Settings in request body.

    Parameters
    ----------
    type : str
        The interval, either "Days", "Hours", "Minutes"
    value: int
        Increment of the interval

    Example
    -------

    Interval(type="Hours", value=2)

    Is an interval of 2 hours.

    """

    type: INTERVAL_TYPES
    value: int

    def __post_init__(self):
        self.validate()

    def validate(self):
        if self.type not in INTERVAL_TYPES.__args__:
            raise ValueError(
                f"{self.type} not a valid interval-type: {INTERVAL_TYPES.__args__}"
            )

    def json(self):
        self.validate()
        return {snake_to_pascal_case(k): v for k, v in self.__dict__.items()}


@dataclass
class ReaderSettings:
    """WIWB reader-settings

    Parameters
    ----------
    start_date: datetime.date
        Reader start_date
    end_date: datetime.date
        Reader end_date
    variable_codes: List[str]
        List of WIWB variable codes
    interval: Interval
        time-interval for reader
    extent: Extend
        extent for reader
    """

    start_date: date
    end_date: date
    variable_codes: list
    interval: Union[Interval, None] = None
    extent: Union[Extent, None] = field(default_factory=Extent)

    def json(self):
        dict = self.__dict__.copy()
        dict["start_date"] = dict["start_date"].strftime("%Y%m%d%H%M%S")
        dict["end_date"] = dict["end_date"].strftime("%Y%m%d%H%M%S")
        for k in ["interval", "extent"]:
            if dict[k] is None:
                dict.pop(k)
            else:
                dict[k] = dict[k].json()

        return {snake_to_pascal_case(k): v for k, v in dict.items() if v is not None}


@dataclass
class Reader:
    """WIWB reader

    Parameters
    ----------
    data_source_code: str
        WIWB datasourcecode to read
    settings: Settings
        WIWB reader settings
    """

    data_source_code: str
    settings: Union[ReaderSettings, None] = field(default_factory=ReaderSettings)

    def json(self):
        dict = self.__dict__.copy()
        dict["settings"] = dict["settings"].json()

        return {snake_to_pascal_case(k): v for k, v in dict.items()}


@dataclass
class ExporterSettings:
    """WIWB export settings

    Parameters
    ----------
    export_projection_file: bool, optional
        To write a projection file (in case of ASCII Grid). Default is False

    """

    export_projection_file = False


@dataclass
class Exporter:
    f"""WIWB exporter

    Parameters
    ----------
    data_format_code: {DATA_FORMAT_CODES}, optional
        data-format code to export data to. Defaults to geotiff
    settings: ExporterSettings
        WIWB exporter settings

    """

    data_format_code: DATA_FORMAT_CODES = "geotiff"
    settings: Union[ExporterSettings, None] = None

    def __post_init__(self):
        self.validate()

    def validate(self):
        if self.data_format_code not in DATA_FORMAT_CODES.__args__:
            raise ValueError(
                f"{self.data_format_code} not a valid data-format-code: {DATA_FORMAT_CODES.__args__}"
            )

    def json(self):
        self.validate()
        return {
            snake_to_pascal_case(k): v
            for k, v in self.__dict__.items()
            if v is not None
        }


@dataclass
class RequestBody:
    """GetGrids request Body"""

    readers: List[Reader]
    exporter: Exporter

    def json(self):
        dict = self.__dict__.copy()
        dict["readers"] = [i.json() for i in dict["readers"]]
        dict["exporter"] = dict["exporter"].json()
        return {snake_to_pascal_case(k): v for k, v in dict.items()}


@dataclass
class GetGrids(Request):
    """GetGrids request"""

    data_source_code: str
    variable_code: str
    start_date: date
    end_date: date
    unzip: bool = True
    interval: Tuple[str, int] = ("Hours", 1)
    data_format_code: DATA_FORMAT_CODES = "geotiff"
    geometries: InitVar[
        GeoSeries | Iterable[Union[Point, Polygon, MultiPolygon]] | None
    ] = None
    bounds: InitVar[Union[Tuple[float, float, float, float], None]] = defaults.bounds

    _response: Union[requests.Response, None] = field(
        init=False, default=None, repr=False
    )
    _geoseries: int = field(init=False, default=None)
    _bounds: Union[Tuple[float, float, float, float], None] = field(
        init=False, default=None
    )

    def __post_init__(self, geometries, bounds):
        self.set_geometries(geometries)
        self.set_bounds(bounds)

    @property
    def epsg(self):
        return defaults.crs

    @property
    def crs(self):
        return self.body.readers[0].settings.extent.crs

    @property
    def body(self) -> RequestBody:
        reader_settings = ReaderSettings(
            self.start_date,
            self.end_date,
            [self.variable_code],
            interval=Interval(*self.interval),
            extent=Extent(*self.bounds),
        )

        reader = Reader(self.data_source_code, settings=reader_settings)

        exporter = Exporter(data_format_code=self.data_format_code)

        return RequestBody(readers=[reader], exporter=exporter)

    @property
    def bbox(self):  # noqa:F811
        return self._bounds

    @property
    def file_name(self):
        stem = "_".join(
            [
                self.data_source_code,
                self.variable_code,
                self.start_date.isoformat(),
                self.end_date.isoformat(),
            ]
        )
        suffix = FILE_SUFFICES[self.data_format_code]
        return f"{stem}.{suffix}"

    @property
    def geoseries(self) -> GeoSeries:
        return self._geoseries

    @property
    def url_post_fix(self) -> str:
        return "grids/get"

    def _to_geoseries(
        self,
        geometries: Optional[GeoSeries | Iterable[Union[Point, Polygon, MultiPolygon]]],
    ) -> GeoSeries:

        # convert iterable to GeoSeries
        if geometries is not None:
            if not isinstance(geometries, GeoSeries):
                geometries = GeoSeries(geometries)

            # Check if geometries are Point, Polygon, or MultiPolygon
            if not all(
                (
                    i in ["Point", "Polygon", "MultiPolygon"]
                    for i in geometries.geom_type
                )
            ):
                raise ValueError(
                    f"Geometries must be Point, Polygon, or MultiPolygon, got {geometries.geom_type.unique()}"
                )

        geometries = self._reproject_geoseries(geoseries=geometries)
        return geometries

    def _reproject_geoseries(self, geoseries: GeoSeries) -> GeoSeries:
        """Set or reproject geoseries to self.epsg"""
        if geoseries.crs is None:
            logger.warning(f"no crs specified in geoseries, will be set to {self.epsg}")
            geoseries.crs = self.epsg
        else:
            geoseries = geoseries.to_crs(self.epsg)
        return geoseries

    def _get_bounds(self, bounds: Union[Tuple[float, float, float, float], None]):
        if (
            self._geoseries is not None
        ):  # if geometries are specified, we'll get bounds from geometries
            bounds = tuple(self._geoseries.total_bounds)
            if bounds is None:
                logger.warning(
                    "bounds will be ignored as long as geometries are not None"
                )
        elif bounds is None:  # if geometries aren't specified, user has to set bounds
            raise ValueError(
                """Specify either 'geometries' or 'bounds', both are None"""
            )
        return bounds

    def run(self):
        self._response = None
        self._response = requests.post(
            self.url, headers=self.auth.headers, json=self.body.json()
        )

        if not self._response.ok:
            self._response.raise_for_status()

    def set_geometries(
        self,
        geometries: Optional[GeoSeries | Iterable[Union[Point, Polygon, MultiPolygon]]],
    ) -> None:
        """Set a list or GeoSeries with Point, Polygon or MultiPolygon values. Handles conversion to
        GeoSeries and reprojection

        Parameters
        ----------
        geometries : GeoSeries | Iterable[Union[Point, Polygon, MultiPolygon]]
            A list or GeoSeries with Point, Polygon and Multipolygon objects
        """
        if geometries is not None:
            geoseries = self._to_geoseries(geometries)
            self._geoseries = geoseries
        else:
            self._geoseries = geometries

    def set_bounds(self, bounds: Tuple[float, float, float, float]) -> None:
        """Set new bounds values. Fits bounds to geoseries.bounds

        Parameters
        ----------
        bounds : Tuple[float, float, float, float]
            Bounds tuple

        """

        bounds = self._get_bounds(bounds)
        self._bounds = bounds

    def write_tempfile(self):
        with tempfile.NamedTemporaryFile(
            suffix=FILE_SUFFICES[self.data_format_code], delete=False
        ) as tmp_file:
            tmp_file_path = Path(tmp_file.name)
            tmp_file.write(self._response.content)
        return tmp_file_path

    def sample(self, stats: str | List[str] = "mean") -> DataFrame:
        """Sample statistics per geometry

        Parameters
        ----------
        stats : str | List[str]
            statistics to sample, provided as list of statistics or a string with one statistic. defaults to mean

            All stats in rasterstats.zonal_stats are available: https://pythonhosted.org/rasterstats/manual.html#statistics
            Common values are:
                - mean: average value of all cells in polygon
                - max: maximum value of all cells in polygon
                - min: minimum value of all cells in polygon
                - percentile_#: percentile value of all cells in polygon. E.g. percentile_50, gives 50th percentile (median) value

            Notes:
            - Providing multiple values, will create a multi-index column in your dataframe
            - Providing multiple statistics, as specified above, doesn't make much sense as it will always return the same value
        """  # noqa:E501

        # check if geometries are set
        if self._geoseries is None:
            raise TypeError(
                """'geometries' is None, should be list or GeoSeries. Set it first"""
            )

        # check if data_format_code is netcdf
        if self.data_format_code != "netcdf4.cf1p6":
            self.data_format_code = "netcdf4.cf1p6"
            self.run()

        # re-run
        if self._response is None:
            self.run()

        # write content in temp-file
        temp_file = self.write_tempfile()

        # sample temp_file
        df = sample_netcdf(
            nc_file=temp_file,
            variable_code=self.variable_code,
            geometries=self.geoseries,
            stats=stats,
            unlink=True,
        )

        return df

    def to_directory(self, output_dir: Union[str, Path]):
        """Write response.content to an output-file"""
        if self._response is None:
            self.run()

        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)
        output_file = output_dir / self.file_name
        output_file.write_bytes(self._response.content)



from wiwb.api_calls.base import Request
from wiwb.api_calls.get_data_sources import GetDataSources
from wiwb.api_calls.get_variables import GetVariables
from wiwb.api_calls.get_grids import GetGrids

__all__ = ["Request", "GetDataSources", "GetVariables", "GetGrids"]



# %%
from shapely.geometry import Point, Polygon, MultiPolygon, box
from typing import Literal
from geopandas import GeoSeries
import os

API_URL = "https://wiwb.hydronet.com/api"
AUTH_URL = (
    "https://login.hydronet.com/auth/realms/hydronet/protocol/openid-connect/token"
)

CLIENT_ID = os.getenv("wiwb_client_id")
CLIENT_SECRET = os.getenv("wiwb_client_secret")

FILE_SUFFICES = {
    "geotiff": "zip",
    "aaigrid": "hdf5",
    "hdf5": "hdf5",
    "netcdf4.cf1p6": "nc",
    "netcdf4.cf1p6.zip": "zip",
}

PRIMARY_STRUCTURE_TYPES = Literal[
    "EnsembleGrid",
    "EnsembleTimeSeries",
    "Event",
    "Grid",
    "ModelGrid",
    "ModelTimeSeries",
    "TimeSeries",
]

DATA_FORMAT_CODES = Literal[*FILE_SUFFICES.keys()]

IMPLEMENTED_GEOMETRY_TYPES = [Point, Polygon, MultiPolygon]

INTERVAL_TYPES = Literal["Days", "Hours", "Minutes"]

CRS_EPSG = 28992
LL_POINT = Point(119865, 449665)
UR_POINT = Point(127325, 453565)
OTHER_POINT = Point(135125, 453394)
POLYGON = box(LL_POINT.x, LL_POINT.y, UR_POINT.x, UR_POINT.y)
GEOSERIES = GeoSeries(
    [LL_POINT, UR_POINT, OTHER_POINT, POLYGON],
    index=["ll_point", "ur_point", "other_point", "polygon"],
    crs=CRS_EPSG,
)


class Defaults:
    bounds: tuple[float, float, float, float] = (109950, 438940, 169430, 467600)
    crs: int = 28992
    geoseries = GEOSERIES


def get_defaults(**kwargs):
    return Defaults(**kwargs)


# %%



"""For converting one thing into another"""


def snake_to_pascal_case(snake_case: str) -> str:
    """Convert snake_case to PascalCase."""
    words = snake_case.split("_")
    return "".join(i.title() for i in words)



# %%
from wiwb.api_calls import GetGrids
from datetime import date
import pytest


def test_grids(auth, api, tmp_path, geoseries, grids_df):
    """
    note WIWB does guarantee you get all data within bounds. Therefore other_point will be outside
    bounds and result is None :-(

    """
    grids = GetGrids(
        auth=auth,
        base_url=api.base_url,
        data_source_code="Meteobase.Precipitation",
        variable_code="P",
        start_date=date(2018, 1, 1),
        end_date=date(2018, 1, 2),
        data_format_code="netcdf4.cf1p6",
        geometries=geoseries,
    )

    df = grids.sample()

    # check if we have results and if they are as expected
    assert not df.empty
    assert (df * 100).astype(int).equals(grids_df)

    # check write to disck
    grids.to_directory(tmp_path)
    assert tmp_path.joinpath(
        "Meteobase.Precipitation_P_2018-01-01_2018-01-02.nc"
    ).exists()


def test_reproject(auth, api, geoseries):
    """
    note WIWB does guarantee you get all data within bounds. Therefore other_point will be outside
    bounds and result is None :-(

    """
    geoseries = geoseries.copy()

    # assume we provide in lat-lon
    geoseries = geoseries.to_crs(4326)
    assert geoseries.crs.to_epsg() == 4326

    # see if geoseries are reprojected at init
    grids = GetGrids(
        auth=auth,
        base_url=api.base_url,
        data_source_code="Meteobase.Precipitation",
        variable_code="P",
        start_date=date(2018, 1, 1),
        end_date=date(2018, 1, 2),
        data_format_code="netcdf4.cf1p6",
        geometries=geoseries,
    )

    assert grids.geoseries.crs.to_epsg() == 28992


def test_bounds(auth, api, defaults):
    # init grids without bounds, should result in default bounds
    grids = GetGrids(
        auth=auth,
        base_url=api.base_url,
        data_source_code="Meteobase.Precipitation",
        variable_code="P",
        start_date=date(2018, 1, 1),
        end_date=date(2018, 1, 2),
        data_format_code="netcdf4.cf1p6",
    )

    assert grids.bbox == defaults.bounds

    # setting bounds to None should result in ValueError ("Specify either 'geometries' or 'bounds', both are None")
    with pytest.raises(
        ValueError, match="Specify either 'geometries' or 'bounds', both are None"
    ):
        grids.set_bounds(None)

    # setting geometries
    grids.set_geometries(defaults.geoseries)
    grids.set_bounds(None)

    assert grids.bbox == tuple(defaults.geoseries.total_bounds)



from wiwb.auth import Auth
from dataclasses import dataclass
from typing import Union
from wiwb.api_calls.get_data_sources import GetDataSources
from wiwb.api_calls.get_variables import GetVariables
from wiwb.api_calls.get_grids import GetGrids
from wiwb.constants import API_URL


@dataclass
class Api:
    """Python API for WIWB service."""

    auth: Union[Auth, None] = None
    base_url: str = API_URL

    def __post_init__(self):
        if self.auth is None:
            self.auth = Auth()
        if self.base_url is None:
            raise ValueError(
                f"Provide a valid base_url. Current value is {self.base_url}"
            )

    def get_data_sources(self, **kwargs):
        api_call = GetDataSources(base_url=self.base_url, auth=self.auth, **kwargs)
        return api_call.run()

    def get_variables(self, **kwargs):
        api_call = GetVariables(base_url=self.base_url, auth=self.auth, **kwargs)
        return api_call.run()

    def get_grids(self, **kwargs):
        get_grids = GetGrids(base_url=self.base_url, auth=self.auth, **kwargs)
        return get_grids



from dataclasses import dataclass
from wiwb.auth import Auth


@dataclass
class Request:
    auth: Auth
    base_url: str

    def __post_init__(self):
        if self.base_url.endswith("/"):
            self.base_url = self.base_url[:-1]

    @property
    def url(self) -> str:
        return f"{self.base_url}/{self.url_post_fix}"
